/*
 * Copyright (c) 2021 CriticalBlue Ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
 * documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

import 'dart:async';
import 'dart:core';

import 'package:flutter/cupertino.dart';
import 'package:flutter/services.dart';
import 'package:logger/logger.dart';
import 'package:shared_preferences/shared_preferences.dart';


// logger
final Logger Log = Logger();

/// Minimal basic results from an Approov token fetch
class TokenFetchResult {
  // Indicates whether a new configuration is available from fetchConfig()
  bool isConfigChanged = false;
  
  // Loggable Approov token string
  String loggableToken = "";

  // Convenience constructor
  TokenFetchResult.fromTokenFetchResultMap(Map aTokenFetchResultMap) {
    isConfigChanged = aTokenFetchResultMap["IsConfigChanged"];
    loggableToken = aTokenFetchResultMap["LoggableToken"];
  }
}

// Approov Discovery to simply initialize the SDK and fetch a single Approov token
class ApproovDiscovery {
  static const MethodChannel _channel = const MethodChannel('approov_discovery');

  /// Initialize the Approov SDK. This must be called prior to any other methods on the Approov SDK. The SDK is
  /// initialized with an initial configuration and an optional update configuration and comment. The initial
  /// configuration is provided from the call in the user app.
  ///
  /// @param initialConfig is the initial configuration string and must be present
  /// @param dynamicConfig is any dynamic update configuration string or null if there is none
  /// @param reserved is provided for future usage
  /// @throws Exception if the provided configuration is not valid
  static Future<void> initialize(String initialConfig, String? dynamicConfig, String? reserved) async {
    final Map<String, dynamic> arguments = <String, dynamic>{
      "initialConfig": initialConfig,
      "dynamicConfig": dynamicConfig,
      "comment": reserved,
    };
    try {
      await _channel.invokeMethod('initialize', arguments);
      isInitialized = true;
    } catch (err) {
      throw Exception('$err');
    }
  }
  
  /// Initiates a request to obtain an Approov token. All calls must provide a URL which provides the high level domain
  /// of the API but for this case only a dummy URL needs is passed.
  ///
  /// @return results of fetching a token
  static Future<TokenFetchResult> fetchApproovToken() async {
    final Map<String, dynamic> arguments = <String, dynamic>{
      "url": "www.approov.io",
    };
    try {
      Map tokenFetchResultMap = await _channel.invokeMethod('fetchApproovTokenAndWait', arguments);
      TokenFetchResult tokenFetchResult = TokenFetchResult.fromTokenFetchResultMap(tokenFetchResultMap);
      return tokenFetchResult;
    } catch (err) {
      throw Exception('$err');
    }
  }

  /// Fetches the current configuration for the SDK. This is done after a notification for an update.
  ///
  /// @return String representation of the configuration.
  static Future<String> fetchConfig() async {
    try {
      String config = await _channel.invokeMethod('fetchConfig');
      return config;
    } catch (err) {
      throw Exception('$err');
    }
  }

  // logging tag
  static const String TAG = "ApproovDiscovery";

  // keys for the Approov shared preferences
  static const String APPROOV_CONFIG = "approov-config";
  static const String APPROOV_PREFS = "approov-prefs";

  // indicates whether ApproovDiscovery has been initialized
  static bool isInitialized = false;

  // config string used to initialize the SDK
  static String initConfigString = "";

  /// Starts the ApproovDiscovery integration.
  ///
  /// @param configString is the initial config string providing account information
  static Future<void> start(String configString) async {
    // ignore multiple attempts to initialize the SDK but log if the initial config chaages
    if (isInitialized) {
      if (configString != initConfigString)
        Log.e("$TAG: Approov Reinitialized with different config $configString, was $initConfigString");
      return;
    }

    // widgets must be initialized for native code support
    WidgetsFlutterBinding.ensureInitialized();

    // initialize the Approov SDK
    String? dynamicConfig = await getApproovDynamicConfig();
    try {
      await initialize(configString, dynamicConfig, null);
      isInitialized = true;
      initConfigString = configString;
    } catch (e) {
      Log.e("$TAG:  Approov initialization failed: ${e.toString()}");
      return;
    }

    // fetch an Approov token to a fixed domain - this will likely result in UNKNOWN_URL but will cause device data to be
    // transmitted to the Approov servers
    TokenFetchResult approovResults = await fetchApproovToken();
    Log.i("$TAG: Approov Discovery Startup: ${approovResults.loggableToken}");

    // update the dynamic configuration if required
    if (approovResults.isConfigChanged) {
      Log.i("$TAG: Approov dynamic configuration updated");
      putApproovDynamicConfig(await fetchConfig());
    }
  }

  /// Stores an application's dynamic configuration string in non-volatile storage.
  ///
  /// The default implementation stores the string in shared preferences, and setting the config string to null is
  /// equivalent to removing the config.
  ///
  /// @param config a config string
  static void putApproovDynamicConfig(String config) {
    SharedPreferences.getInstance().then((prefs) {
      prefs.setString(APPROOV_CONFIG, config);
    });
  }

  /// Returns the application's dynamic configuration string from non-volatile storage.
  ///
  /// The default implementation retrieves the string from shared preferences.
  ///
  /// @return config string, or null if not present
  static Future<String?> getApproovDynamicConfig() async {
    SharedPreferences prefs = await SharedPreferences.getInstance();
    String? dynamicConfig = prefs.getString(APPROOV_CONFIG);
    return dynamicConfig;
  }
}
